import json
from time import perf_counter

from .log import log

# Pull environment variables from .env or os
from dotenv import load_dotenv
load_dotenv()

# disable warning for debugging of frozen packages
# seems to only matter when running under Jupyter
import os
os.environ["PYDEVD_DISABLE_FILE_VALIDATION"]="1"

import papermill as pm

class Workflow:

    def __init__(self, workflow: dict):
        self.workflow=workflow
        self.verify()

    @classmethod
    def from_json(cls,j:str):
        wf:Workflow = cls(json.loads(j))
        return wf
        
    @classmethod
    def from_file(cls,filepath:str):
        j=json.load(open(filepath))
        return cls(j)

    @classmethod
    def from_vars(cls,output_folder:str, notebook_order:list[str], variables:dict):
        workflow = {}
        workflow["Output"] = output_folder
        nblist = []
        for notebook in notebook_order:
            nblist.append([notebook, variables])
        workflow["Notebooks"] = nblist
        return cls(workflow)

    def verify(self):
        '''
        Walk thru the structure of the workflow and ensure it is setup correctly.
        '''
        has_errors:bool=False
        errors:str=""

        # Verify Output
        path=self.workflow["Output"]
        if path is None:
            log.error("'Output' path is set to None.")
            has_errors=True
        elif not os.path.exists(path):
            log.info("'Output' path ({path}) does not exist, creating it...".format(path=path))
            self.ensure_path(path) 
        elif not os.path.isdir(path):
            log.error("'Output' path ({path}) is not a directory.".format(path=path))  
            has_errors=True 
        elif not os.access(path, os.W_OK):
            log.error("'Output' path ({path}) is not writable.".format(path=path))
            has_errors=True

        log.debug("Output config verified.")

        # Verify Common Settings
        if "Common" in self.workflow:
            common_dict=self.workflow["Common"]
            log.debug("Common config exists.")    

        # Verify Notebooks    
        nblist=self.workflow["Notebooks"]
        if nblist is None:
            log.error("'Notebooks' is set to None.")
            has_errors=True

        c = 1
        for step in nblist:
            notebook:str=step[0]
            if not os.path.isfile(notebook):
                log.error("'Notebooks' #{c}: '{notebook}' not exist.".format(c=c,notebook=notebook))
                has_errors=True
            elif not os.access(notebook,os.R_OK):
                log.error("'Notebooks' #{c}: '{notebook}' is not readable.".format(c=c,notebook=notebook))  
                has_errors=True

            params=step[1]
            if params is None:
                log.error("'Notebooks' #{c}: '{notebook}' does not have parameters set.".format(c=c,notebook=notebook))
                has_errors=True
            p=1
            for key, value in params.items():
                if not type(key) == str:
                    log.error("'Notebooks' #{c}: {notebook} Variable #{p}: name '{var}' is not a string value.".format(c=c,notebook=notebook,p=p,var=key))
                    has_errors=True
                p=p+1
            c=c+1
        
        if has_errors:
            raise ValueError("Error in Workflow specification (see errors in log).")
        else:
            log.debug("No errors in verification of workflow specification, exporting a copy to output directory.")
            self.export_workflow(os.path.join(self.workflow["Output"],"workflow.json"))

    def export_workflow(self, filepath:str) ->json:
        with open(filepath, 'w', encoding='utf-8') as f:
            json.dump(self.workflow, f, ensure_ascii=False, indent=4)

    def ensure_path(self,path):
        '''
        Ensure_path(path) simply creates a directory if it is not there already.
        '''
        if not os.path.exists(path):
            os.makedirs(path)
            log.info("Created output directory: "+path)

    def execute_notebook(self, notebook:str, variables:dict, output_path:str) -> float:
        '''
        Execute_notebook runs a single notebook file in PaperMill returns the run time in seconds
        '''
        try:
            out_notebook:str = os.path.join(output_path,notebook.replace('/',"_"))
            log.debug("Now executing notebook '{notebook}' with params '{params}' and output '{output}'.".format(notebook=notebook,params=variables,output=output_path))

            notebook_working_dir = os.path.dirname(notebook)
            if notebook_working_dir == '': # no directory given
                notebook_working_dir=os.getcwd()

            log.debug("CWD for notebook '{notebook}' set to '{cwd}'.".format(notebook=notebook,cwd=notebook_working_dir))

            step_start=perf_counter()
            pm.execute_notebook(notebook, out_notebook, variables, cwd=notebook_working_dir)
            step_end=perf_counter()
            return (step_end-step_start)
        
        except Exception as e:
            log.error("There was an error executing the notebook ({nb}), please verify the parameters and that the notebook runs successfully on its own.".format(nb=notebook))
            raise e

    def run(self):
        '''
        Run the defined workflow files in order in PaperMill
        '''
        output_path=self.workflow["Output"]
        self.ensure_path(output_path)

        common_params = None
        if "Common" in self.workflow:
            common_params=self.workflow["Common"]

        workflow_start=perf_counter()
        c=1
        for step in self.workflow["Notebooks"]:
            notebook:str=step[0]
            params=step[1]

            if common_params is not None:
                params = params | common_params
                log.debug("Common config exists, combined to {}".format(params)) 


            print("\nStarting Notebook #{c}: {name}".format(c=c,name=notebook))
            time=self.execute_notebook(notebook,params,output_path)
            print("Completed #{c}: {name} in {time} seconds.".format(c=c,name=notebook,time=round(time,2)))

            c=c+1
        workflow_end=perf_counter()
        print("Workflow completed successfully in {time} seconds.".format(time=round(workflow_end-workflow_start,2)))




import sys

import pytest
from neo4j.exceptions import ClientError

sys.path.insert(0, 'src')

from opentldr import Workflow
from opentldr.Domain import ReferenceNode


NOTEBOOK_DIR = "tests"

def test_create_missing_output():
    with pytest.raises(Exception):
        wf = Workflow({})


def test_create_missing_notebooks():
    with pytest.raises(Exception):
        wf = Workflow({
            "Output": "./temp",
        })


def test_create_missing_vars():
    with pytest.raises(Exception):
        wf = Workflow({
            "Output": "./temp",
            "Notebooks": [f"{NOTEBOOK_DIR}/Step_0.ipynb"],
        })


def test_create_1_step():
    wf = Workflow({
        "Output": "./temp",
        "Notebooks": [(f"{NOTEBOOK_DIR}/Step_0.ipynb", {})],
    })


def test_create_2_steps():
    wf = Workflow({
        "Output": "./temp",
        "Notebooks": [
            (f"{NOTEBOOK_DIR}/Step_0.ipynb", {}),
            (f"{NOTEBOOK_DIR}/Step_1.ipynb", {}),
        ],
    })






import sys

import pytest
from neo4j.exceptions import ClientError

sys.path.insert(0, 'src')

from opentldr import KnowledgeGraph
from opentldr.Domain import ReferenceNode


def test_default_connection():
    kg = KnowledgeGraph()
    assert kg.connect()


def test_run_query():
    kg = KnowledgeGraph()

    rn = kg.add_reference_node(ReferenceNode(text="this is a test"), type="TEST")
    list = kg.cypher_query("MATCH (n:ReferenceNode) return n", "n")
    assert rn in list

    kg.delete_reference_node(rn)
    list = kg.cypher_query("MATCH (n:ReferenceNode) return n", "n")
    assert rn not in list




def test_run_query_params():
    kg = KnowledgeGraph()

    rn = kg.add_reference_node(ReferenceNode(text="this is a test"), type="TEST")
    uid = rn.uid
    list = kg.cypher_query("MATCH (n:ReferenceNode) WHERE n.uid = $uid RETURN n", "n",
                           params={"uid": uid})
    assert rn in list

    kg.delete_reference_node(rn)
    list = kg.cypher_query("MATCH (n:ReferenceNode) WHERE n.uid = $uid RETURN n", "n",
                           params={"uid": uid})
    assert rn not in list


def test_run_query_missing_params():
    kg = KnowledgeGraph()

    rn = kg.add_reference_node(ReferenceNode(text="this is a test"), type="TEST")
    with pytest.raises(ClientError):
        list = kg.cypher_query("MATCH (n:ReferenceNode) WHERE n.uid = $uid RETURN n", "n",
                               params={})           # uid is NOT defined in params




import os
import re
from abc import ABC, abstractmethod 
from datetime import datetime

from dotenv import load_dotenv
load_dotenv()

from .KnowledgeGraph import KnowledgeGraph

from .AbstractDataRepo import AbstractDataRepo
from .FileSystemDataRepo import FileSystemDataRepo
from .S3DataRepo import S3DataRepo

from .log import log

class DataRepo(AbstractDataRepo):
    '''
    Factory class to setup any type of DataRepo from the configuration inputs.
    '''
    repo:AbstractDataRepo


    def _configOrEnv(self, key:str, config:dict) -> str:
        value:str=config.get(key)

        # setting in environment to override configurations for automation
        if os.getenv("OVERRIDE_CONTENT_REPO_CONFIGS") is not None:
            value=None

        if value is None:
            value = config.get(key.capitalize())
        if value is None:
            value = os.getenv(key)
        if value is None:
            value = os.getenv(key.capitalize())
        if value is None:
            raise TypeError("No value found for '{key}', which is required.".format(key=key))
        return value


    def __init__(self, kg:KnowledgeGraph, config:dict):
        if config is None:
            log.warn("No config file passed to DataRepo, defaulting to using environment variables.")
            config="{}"

        type = self._configOrEnv("repo_type",config);
        # TODO: The s3/files check here make the "case _" branch unreachable.
        # I recommend deleting the check here.  Making it easier to expand cases below in the future.
        if type is None or type not in ['s3','files']:
            message=("DataRepo factory requires a dictionary structure as a parameter that includes at least a 'repo_type' entry set to one of 's3' or 'files'.")
            log.error(message)
            raise TypeError(message)
        
        match type:
            case "s3":
                bucket_name=self._configOrEnv("bucket",config)
                aws_access_key_id=self._configOrEnv("aws_access_key_id",config)
                aws_secret_access_key=self._configOrEnv("aws_secret_access_key",config)
                prefix=self._configOrEnv("prefix",config)
                self.repo= S3DataRepo(kg,bucket_name=bucket_name, aws_access_key_id=aws_access_key_id, 
                                      aws_secret_access_key=aws_secret_access_key, prefix=prefix)
            case "files":
                ingest_path=self._configOrEnv("path",config)
                self.repo= FileSystemDataRepo(kg,ingest_path=ingest_path)
            case _:
                message:str = "Invalid 'type' config ('{type}').".format(type=type)
                log.error(message)
                raise NotImplementedError(message)

    def importData(self) -> list[str]:
        if self.repo is None:
            log.error("No DataRepo was configured.")
        return self.repo.importData()

    def describe(self) -> str:
        if self.repo is None:
            log.error("No DataRepo was configured.")
        return self.repo.describe()
    
 


from .DataRepo import AbstractDataRepo
from .KnowledgeGraph import KnowledgeGraph
from .Domain import jsonToKg, dictToKg
from .log import log

import boto3
import json
import os
import tempfile

class S3DataRepo(AbstractDataRepo):

    files:list = []

    def __init__(self, kg:KnowledgeGraph, bucket_name:str, aws_access_key_id:str, aws_secret_access_key:str, prefix:str=None) :
        self.kg=kg
        session = boto3.Session( aws_access_key_id, aws_secret_access_key)
        s3 = session.resource('s3')
        self.bucket_name = bucket_name
        self.bucket = s3.Bucket(bucket_name)
        self.prefix = prefix
        for o in self.bucket.objects.all():
            if prefix is None or o.key.startswith(prefix):
                if not o.key.endswith("/"): #skip directories
                    self.files.append(o)


    def describe(self) -> str:
        return "S3 Bucket Content ('{bucket}')".format(bucket=self.bucket_name)

    # TODO: Move repeated method to superclass
    def _importByClass(self,raw:dict,clazz:str) -> list[str]:
        list_of_uids=[]
        if clazz in raw:
            for item in raw[clazz]:
                item['class']=clazz
                o = dictToKg(self.kg,item)
                list_of_uids.append(o.uid)
        return list_of_uids
  

    def importData(self) -> list[str]:
        log.info("Importing Data from S3 Bucket...")
        default_source_name = "S3 Bucket:{path}".format(path=self.bucket_name)
        list_of_uids = []

        # load objects by dependency order but each file only once
        imported=[]
        for c in [*self.import_order, ""]:
            if c !="":
                log.debug("Finding objects for type {type}.".format(type=c))
            else:
                log.debug("Finding remaining files of unknown class types.")

            for summary in self.files: 
                if summary.key not in imported:
                    if c == "" or c in summary.key or c.lower() in summary.key or c.capitalize() in summary.key:
                        imported.append(summary.key)
                        extension = "."+summary.key.rsplit(".",1)[-1]
                        url="https://{bucket_name}.s3.amazonaws.com/{object_name}".format(bucket_name=self.bucket_name, object_name=summary.key)
                        text=summary.get()['Body'].read().decode()
                        
                        match extension:
                            case '.txt' | '.text':
                                log.info("Importing Text Object: {path}".format(path=summary.key))
                                raw=text.split('\n---\n')
                                for part in raw:
                                    ann, text = self.parseAnnotatedText(part)
                                    if not 'class' in ann:
                                        ann['class'] = self.inferDomainClass(summary.key,ann)
                                    if len(text) > 0 and not 'text' in ann:
                                        ann['text']=text
                                    o = dictToKg(self.kg,ann)
                                    list_of_uids.append(o.uid)
                            
                            case '.cql' | '.cypher':
                                log.info("Importing Cypher Object: {path}".format(path=summary.key))
                                with tempfile.NamedTemporaryFile(delete=False) as fp:
                                    fp.write(text.encode())  # this should be cypher
                                    fp.close()      # don't delete on close
                                    self.kg.cypher_import(fp.name)
                                    # deletes when exiting block
                                list_of_uids.extend(self.clean_up_nodes())
                            
                            case '.json':
                                log.info("Importing JSON Object: {path}".format(path=summary.key))
                                raw:dict= json.loads(text)
                                for c in self.import_order:
                                    added_list=self._importByClass(raw,c)
                                    list_of_uids.extend(added_list)
                            
                            case _:
                                log.warning("Skipping unknown-type file: {path}".format(path=summary.key))
                                continue
                            
        return list_of_uids



from .Domain import *
#from .KnowledgeGraph import KnowledgeGraph

import neomodel as nm

# Document Content

class OrderedEdges(nm.StructuredRel):
    order = nm.IntegerProperty(required=False, default=0)

class SubsectionOf(OpenTldrMeta, OpenTldrEdge, OrderedEdges):
    pass

class AuthoredBy(nm.StructuredRel, OpenTldrMeta, OpenTldrEdge, Uncertain):
    pass

class Enriches(nm.StructuredRel, OpenTldrMeta, OpenTldrEdge):
    pass
   
class Enrichment(nm.StructuredNode, OpenTldrMeta, OpenTldrNode):
    contains = nm.RelationshipTo('Enrichment','CONTAINS', model=SubsectionOf, cardinality=ZeroOrMore)

class TechnicalPaper(Enrichment):
    title = nm.StringProperty(required=False)
    link = nm.StringProperty(required=False)
    publish_date = nm.DateProperty(required=False)
    author = nm.RelationshipTo(ReferenceNode, 'AUTHORED_BY', model=AuthoredBy, cardinality=ZeroOrMore)
    enriches = nm.RelationshipTo(Content, 'Enriches', model=Enriches, cardinality=ZeroOrOne)

class Section(Enrichment):
    title = nm.StringProperty(required=True)
    url = nm.StringProperty(required=False)

class TextChunk(Entity, Enrichment):
    index = nm.IntegerProperty(required=False)

class Table(Enrichment):
    pass


class Formula(Enrichment):
    pass


class Figure(Enrichment):
    caption = nm.StringProperty(required=False)
    title = nm.StringProperty(required=True)
    description = nm.StringProperty(required=False)
    url = nm.StringProperty(required=False)

# Image Content

class Image(Enrichment):
    title = nm.StringProperty(required=True)
    description = nm.StringProperty(required=False)
    url = nm.StringProperty(required=False)

class ImageSequence(Enrichment):
    pass


# Audio Content

class AudioChunk(Enrichment):
    pass


class AudioSequence(Enrichment):
    pass


# Video Content

class Scene():
    time:int = 0
    frame:Image = None
    audio:AudioChunk = None


class Video(ImageSequence, AudioSequence):
    pass


# Similarity Matrix

class SimilarTo (nm.StructuredRel, OpenTldrMeta, OpenTldrEdge):
    pass

class Similarity (nm.StructuredNode, OpenTldrMeta, OpenTldrNode, Scored):
    similar_to = nm.Relationship(Content,'SIMILAR_TO', model=SimilarTo, cardinality=ZeroOrMore)
    
    def to_text(self) -> str:
        nodes=""
        for node in self.similar_to:
            if len(nodes)!=0:
                nodes=nodes+" AND "
            nodes= nodes + node.title.replace("\n","")
        return "There is a {score} similarity between: {nodes}".format(score=self.score_to_text(),nodes=nodes)
    



from .AbstractDataRepo import AbstractDataRepo
from .Domain import jsonToKg, dictToKg
from .KnowledgeGraph import KnowledgeGraph
from .log import log

import os
import json


class FileSystemDataRepo(AbstractDataRepo):
    '''
    Reads all the text files in the provided directory path. Creates a Source node for the directory and parses the text files.
    This extracts lines with property definitions (e.g., "date: 05/14/1960 08:00PM") and puts the rest into the text body.
    Creates a Content node, for each file and adds it to the knowledge graph.
    '''

    def __init__(self, kg:KnowledgeGraph, ingest_path="."):     # TODO: "." seems like a bad default. "files" would be better.
        self.kg=kg
        self.ingest_path = ingest_path
        self.files=self.getAllFiles(ingest_path)


    def describe(self) -> str:
        return "File system at ('{path}') contains {n} files.".format(path=self.ingest_path, n=len(self.files))  

    @classmethod
    def getAllFiles(cls,current:str) -> list[str]:
        if os.path.exists(current) and os.access(current,os.R_OK):
            if os.path.isdir(current):
                list=[]
                for child in os.listdir(current):
                    full_path=os.path.join(current, child)
                    #print(full_path)
                    list.extend(cls.getAllFiles(full_path))
                return list
            else:
                return [current]
        else:
            if not os.path.exists(current):
                log.debug("Doesn't exist: {name}.".format(name=current))
            
            if not os.access(current,os.R_OK):
                log.warning("Cannot read: {name}.".format(name=current))
            return []    

    # TODO: Move repeated method to superclass
    def _importByClass(self,raw:dict,clazz:str) -> list[str]:
        list_of_uids=[]
        if clazz in raw:
            for item in raw[clazz]:
                item['class']=clazz
                o = dictToKg(self.kg,item)
                list_of_uids.append(o.uid)
        return list_of_uids

    def importData(self) -> list[str]:
        log.info("Importing Data from File System...")
        list_of_uids=[]

        # load files by dependency order but each file only once
        imported=[]
        for c in [*self.import_order, ""]:
            if c !="":
                log.debug("Finding files for type {type}.".format(type=c))
            else:
                log.debug("Finding remaining files of unknown class types.")

            for fullpath in self.files:  
                if fullpath not in imported:
                    if c == "" or c in fullpath or c.lower() in fullpath or c.capitalize() in fullpath:
                        imported.append(fullpath)
                        try:
                            extension = os.path.splitext(fullpath)[-1]
                            match extension:

                                case '.txt' | '.text':
                                    log.info("Importing Text File: {path}".format(path=fullpath))
                                    with open(fullpath) as f:
                                        raw=f.read().split('\n---\n')
                                        for part in raw:
                                            ann, text = self.parseAnnotatedText(part)
                                            if not 'class' in ann:
                                                ann['class'] = self.inferDomainClass(fullpath,ann)
                                            if len(text) > 0 and not 'text' in ann:
                                                ann['text']=text
                                            o = dictToKg(self.kg,ann)
                                            list_of_uids.append(o.uid)
                                
                                case '.cql' | '.cypher':
                                    log.info("Importing Cypher File: {path}".format(path=fullpath))
                                    self.kg.cypher_import(fullpath)
                                    list_of_uids.extend(self.clean_up_nodes())
                                
                                case '.json':
                                    log.info("Importing JSON File: {path}".format(path=fullpath))
                                    with open(fullpath) as f:
                                        raw:dict= json.load(f)

                                        for c2 in self.import_order:
                                            added_list=self._importByClass(raw,c2)
                                            list_of_uids.extend(added_list)
                                
                                case _:
                                    log.warning("Skipping unknown-type file: {path}".format(path=fullpath))
                                    continue
                        except:
                            log.error("Failed to import file: {}".format(fullpath))
                            
        return list_of_uids



import json
from abc import ABC, abstractmethod 
from datetime import datetime,date

# All of the OpenTLDR Domain classes build on Neomodel as an OGM for Neo4j
import neomodel as nm

# Note change in NeoModel 5.3.0
# from neomodel.cardinality import One, ZeroOrOne, ZeroOrMore, OneOrMore
from neomodel.sync_.cardinality import One, ZeroOrOne, ZeroOrMore, OneOrMore

from .log import log

def jsonToKg(kg,json_string):
    '''
    Hydrate a json string to a Domain object. This requires the "class" attribute to
    determine the correct class type, so it only works on these classes.
    '''
    raw = json.loads(json)
    log.info("Ingesting JSON into an OpenTLDR Domain object and then the KG.")
    print("JSON: ",raw)
    if isinstance(raw,list):
        for item in raw:
            dictToKg(kg,item)
    elif isinstance(raw,dict):
        dictToKg(kg,raw)
    

def dictToKg(kg,data:dict):

    if "class" in data:
        clazz = data["class"]
        data.pop("class")
        obj= globals()[clazz].from_json(kg,json.dumps(data))
        return obj
    else:
        raise TypeError("Cannot detect a class specification in this data. Unable to instantiate object.")


def inferDateFormat(date_string:str) -> date:
    formats=['%m/%d/%y', '%m/%d/%Y', '%m-%d-%y', '%m-%d-%Y', '%Y-%m-%d']
    d:date = None
    for f in formats:
        try:
            d=datetime.strptime(date_string,f)
            return d
        except:
            pass # try try again
    # give up with a nicely formatted date, even if wrong
    log.error("Unable to parse date out of '{date}'.  Returning now.".format(date=date))
    return datetime.now 


# UTILITY MIX-IN CLASSES

class OpenTldrMeta():
    '''
    OpenTldrMeta is a mixin class that adds a unique id (uid) and JSON property (metadata)
    to a Node or Edge. This is used in every Edge and Node in OpenTLDR and provides an
    interface for PyDantic for data manipulation and JSON.
    '''
    uid=nm.UniqueIdProperty()
    metadata=nm.JSONProperty()
       
    def to_text(self) -> str:
        '''
        to_text() returns a unique, user readable string representing the object.
        '''
        return "uid: {uid}".format(uid=self.uid)


class OpenTldrNode():
    '''
    OpenTldrNode is mixed in to each node class to support json import/export.
    '''
    @classmethod
    def from_json(cls, kg, json_string):
        json_dict=json.loads(json_string)
        if "class" in json_dict:
            json_dict.pop("class")
        if "date" in json_dict:
            json_dict["date"]=inferDateFormat(json_dict["date"])
        node = cls(**json_dict)
        node.from_json_connect(kg, json_dict)
        node.save()
        return node
        
    def to_json(self,kg) -> str:
        data = self.__properties__.copy()
        if "date" in data:
            data["date"]=data["date"].strftime("%m/%d/%Y")
        data["class"]=type(self).__name__
        if "element_id_property" in data:
            data.pop("element_id_property")
        connection = self.to_json_connect(kg, data)
        if connection is not None:
            data.update(connection)
        return json.dumps(data)

    def to_json_connect(self, kg, json_dict) -> dict:
        '''
        Override to_json_connect to insert edge data into the node.
        '''
        pass

    def from_json_connect(self, kg, json_dict):
        '''
        Override from_json_connect to rebuild edge data from serialized node.
        '''
        pass



class OpenTldrEdge():
    '''
    OpenTldrEdge is mixed in to each edge class to support json import/export
    '''

    @abstractmethod
    def get_connection_json() -> dict:
        pass

    @classmethod
    def from_json(cls, kg, json_string):
        json_dict=json.loads(json_string)
        if "class" in json_dict:
            json_dict.pop("class")
        if "date" in json_dict:
            json_dict["date"]=inferDateFormat(json_dict["date"])
        node = cls(**json_dict)
        node.from_json_connect(kg, json_dict)
        node.save()
        return node
        
    def to_json(self,kg) -> str:
        data = self.__properties__.copy()
        if "date" in data:
            data["date"]=data["date"].strftime("%m/%d/%Y")
        data["class"]=type(self).__name__
        if "element_id_property" in data:
            data.pop("element_id_property")
        connection = self.to_json_connect(kg, data)
        if connection is not None:
            data.update(connection)
        return json.dumps(data)

    def to_json_connect(self, kg, json_dict) -> dict:
        '''
        Override to_json_connect to insert edge data into the node.
        '''
        pass

    def from_json_connect(self, kg, json_dict):
        '''
        Override from_json_connect to rebuild edge data from serialized node.
        '''
        pass

    @classmethod
    def from_json(cls, kg, json_string):
        json_dict=json.loads(json_string)
        if "connection" in json_dict:
            connection=json_dict["connection"]

            from_node=kg.get_by_uid(connection["start"])
            #print("FROM: {type}\t{uid}\t{desc}".format(type=type(from_node),uid=from_node.uid,desc=from_node.to_text()))
            
            to_node=kg.get_by_uid(connection["end"])
            #print("to: {type}\t{uid}\t{desc}".format(type=type(to_node),uid=to_node.uid,desc=to_node.to_text()))

            rel=getattr(from_node,connection["relation"])
            if "class" in json_dict:
                json_dict.pop("class")
            json_dict.pop("connection")
            edge= rel.connect(to_node,json_dict)
            edge.save()
            return edge
        else:
            raise TypeError("OpenTldrEdge requires connection data stored in json.")

        
    def to_json(self,kg) -> str:
        data = self.__properties__.copy()
        data["class"]=type(self).__name__
        data["connection"]=self.get_connection_json()
        if "element_id_property" in data:
            data.pop("element_id_property")
        return json.dumps(data)


class OpenTldrText():
    '''
    OpenTldrText is a mixin class that adds the properties of text and type (both are Strings)
    to any Node or Edge. This is used whenever the client stores text in the knowledge graph.
    '''
    text = nm.StringProperty(required=True)
    type = nm.StringProperty(required=True)

    def to_text(self) -> str:
        return "type: {type}\t text: {text}".format(type=self.type,text=self.text)


class CitableNode(nm.StructuredNode):
    '''
    CitableNode is a baseclass for any Node from which we extract entities (e.g., NER)
    and we need to refer back to the Node from where they were detected.
    '''
    pass

class Uncertain():
    '''
    Uncertain is a mixin class that adds a confidence value (0.0-1.0) to a Node or Edge.
    This is implemented to store the float value and generate description qualitative
    text clauses for the ranges of this confidence value.
    '''
    confidence = nm.FloatProperty(default=1.0)
    def uncertainty_to_text(self) -> str:
        if self.confidence < 0.05:
            return "with almost no chance"
        if self.confidence >= 0.05 and self.confidence < 0.20:
            return "very unlikely"
        if self.confidence >= 0.2 and self.confidence < 0.45:
            return "unlikely"
        if self.confidence >= 0.45 and self.confidence < 0.55:
            return "with roughly even chance"
        if self.confidence >= 0.55 and self.confidence < 0.8:
            return "likely"
        if self.confidence >= 0.8 and self.confidence < 0.95:
            return "very likely"
        if self.confidence >= 0.55 and self.confidence < 1.0:
            return "with almost certainty"
        return "with certainty"

class Scored():
    '''
    Scored is a mixin class that adds a score value (0.0-1.0) to a Node or Edge.
    This is implemented to store the float value and generate descriptive qualitative
    text clauses for the ranges of this score value.
    '''
    score = nm.FloatProperty(default=-1.0)
    def score_to_text(self) -> str:
        if self.score < 0.0:
            return "unknown"
        if self.score < 0.15:
            return "very low"
        if self.score >= 0.15 and self.score < 0.40:
            return "low"
        if self.score >= 0.4 and self.score < 0.6:
            return "medium"
        if self.score >= 0.6 and self.score < 0.85:
            return "high"
        if self.score >= 0.85 and self.score < 1.0:
            return "very high"
        return "perfect"



# REFERENCE KNOWLEDGE


class ReferenceEdge(nm.StructuredRel, OpenTldrText, OpenTldrMeta, OpenTldrEdge):
    '''
    ReferenceEdge allows for general reference data relationships to be introduced into the 
    knowledge graph and used abstractly (e.g., ontological distance using path finding).
    This uses the OpenTldrText mixin to implement text and type properties.
    '''
    hypothesized = nm.BooleanProperty(default=False)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def is_hypothesized(self) -> bool:
        return self.hypothesized

    def set_hypothesized(self, is_hypothesized:bool):
        self.hypothesized=is_hypothesized

    def get_connection_json(self) -> dict:
        return {
            "start": self.start_node().uid,
            "relation": "edges",
            "end": self.end_node().uid
        }

    def to_text(self) -> str:
        left:ReferenceNode = self.start_node()
        right:ReferenceNode = self.end_node()
        return "{left} -[{relationship}:{type}]-> {right}".format(left=left.text,
                                                           right=right.text,
                                                           relationship=self.text,
                                                           type=self.type)


class ReferenceNode(nm.StructuredNode, OpenTldrText, OpenTldrMeta, OpenTldrNode):
    '''
    ReferenceNode allows for general reference data nodes to be introduced into the
    knowledge graph and used abstractly (e.g., to relate extracted entities to known things).
    This uses the OpenTldrText mixin to implement text and type properties.
    The hypothesized property is a boolean that indicates if this was inferred (e.g., discovered)
    or provided as fact via external reference data (e.g., imported). 
    '''
    hypothesized = nm.BooleanProperty(default=False)

    edges = nm.RelationshipTo('ReferenceNode','REFERENCE_EDGE', model=ReferenceEdge, cardinality=ZeroOrMore)

    def is_hypothesized(self) -> bool:
        return self.hypothesized

    def set_hypothesized(self, is_hypothesized:bool):
        self.hypothesized=is_hypothesized

    def connect_to(self, to_node:'ReferenceNode') -> ReferenceEdge:
        return self.edges.connect(to_node)
    
    def connect(self, to_node:'ReferenceNode', edge:ReferenceEdge):
        self.edges.connect()

    def to_text(self) -> str:
        return "'{text}' of type {type}".format(text=self.text,type=self.type)
    


# Active Data


class Source(nm.StructuredNode, OpenTldrMeta, OpenTldrNode):
    '''
    Source nodes indicate the generator of various content. This may be an author or distributor
    of information. The name property is used to uniquely identify the Source.
    '''
    name = nm.StringProperty(unique_index=True, required=True)


class IsFrom(nm.StructuredRel, OpenTldrMeta, OpenTldrEdge):
    '''
    IsFrom edges connect Content nodes to the Source nodes in the knowledge graph.
    '''
    def to_text(self) -> str:
        return "The {content_type} '{content_title}' is from the source {end}.".format(
            content_type= self.start_node().type,
            content_title=self.start_node().title,
            end=self.end_node().name)


class Content(CitableNode, OpenTldrMeta, OpenTldrText, OpenTldrNode):
    '''
    Content nodes represent text-based information, such as a news article or message.
    They include properties for title (String), date (datetime.date), and url (String).
    In OpenTldr Content node urls are used for hyperlinks back to the original material.
    '''
    title = nm.StringProperty(required=True)
    date = nm.properties.DateProperty(required=True)
    url = nm.StringProperty(required=True)

    is_from = nm.RelationshipTo(Source, 'IS_FROM', model=IsFrom, cardinality=One)

    def get_is_from(self) -> Source:
        return self.is_from.single()

    def to_text(self) -> str:
        return "A {content_type} content titled '{content_title}' from {date}.".format(
            content_type= self.type,
            content_title=self.title,
            date=self.date)
    
    def to_json_connect(self, kg, json_dict) -> dict:
        out:dict= {}
        if "source" not in json_dict:
            out["source"]=self.is_from.single().name
        return out

    def from_json_connect(self, kg, json_dict):
        # Ensure there is an is_from relation
        if "source" in json_dict:
            source=kg.get_source_by_name(json_dict["source"])
            if source is None:
                source=kg.add_source(json_dict["source"])
            self.save()
            self.is_from.connect(source)

class MentionedIn(nm.StructuredRel, OpenTldrMeta, OpenTldrEdge):
    '''
    MentionedIn edges connects an extracted Entity node to the CitableNode (e.g., Content 
    or Request) that included the reference.
    '''
    def to_text(self) -> str:
        return "The entity '{entity}' is mentioned in the '{title}'.".format(
            entity=self.start_node().text,
            title=self.end_node().title)


class RefersTo(nm.StructuredRel, OpenTldrMeta, OpenTldrEdge, Uncertain):
    '''
    RefersTo edges connects an extracted Entity node to the ReferenceNode that it is believed
    to represent with the uncertainity value provided. 
    '''
    def to_text(self) -> str:
        return "The entity '{entity}' {uncertainty} refers to '{ref}'.".format(
            entity=self.start_node().text,
            ref=self.end_node().text,
            uncertainty=self.uncertainty_to_text())


class Entity(nm.StructuredNode, OpenTldrText, OpenTldrMeta, OpenTldrNode):
    '''
    Entity nodes represent specific things that are "mentioned_in" in Citable nodes 
    (e.g., Content or Requests) that "refers_to" some ReferenceNode information. The text property
    contains the text for the entity identified and the type property expresses what it is.
    The "refers_to" edge can be uncertain (e.g., sematic similarity of text and type consistency). 
    '''
    refers_to=nm.RelationshipTo(ReferenceNode,'REFERS_TO', model = RefersTo, cardinality=ZeroOrMore)  
    mentioned_in = nm.RelationshipTo("CitableNode", 'MENTIONED_IN', model=MentionedIn, cardinality=OneOrMore)

    def get_refers_to(self) -> list[Source]:
        return self.refers_to
    
    def get_mentioned_in(self) -> Source:
        return self.mentioned_in.single()


# Information Request


class User(nm.StructuredNode, OpenTldrMeta, OpenTldrNode):
    '''
    User node represents a client of the OpenTLDR system and has name and email properties (Stings).
    Similar to a Content connecting to a Source, Requests connect to a User (see RequestedBy edge).
    '''
    name = nm.StringProperty(required=True)
    email = nm.StringProperty(required=True)

    def to_text(self) -> str:
        return "The User '{name}' with email '{email}'.".format(
            name=self.name, email=self.email)


class RequestedBy(nm.StructuredRel, OpenTldrMeta, OpenTldrEdge):
    '''
    RequestedBy edges connect Request nodes to the User node to reflect who made the request.
    '''

    def to_text(self) -> str:
        return "The request '{request}' was requested by {user}.".format(
            request=self.start_node().title,
            user=self.end_node().name)


class Request(CitableNode, OpenTldrMeta, OpenTldrNode):
    '''
    Request nodes information requests that are made by users to indicate what information they
    are interested in and how to tailor that information for them. A "text" property (String) 
    holds the text of the information request and the "title" property (String) provides a 
    short labels for the request to avoid repeating long request text. The "requested_by" edge
    links each request to one User node.
    '''
    title=nm.StringProperty(required=True)
    text = nm.StringProperty(required=True)

    requested_by = nm.RelationshipTo(User, 'REQUESTED_BY', model=RequestedBy, cardinality=One)

    def get_requested_by(self) -> User:
        return self.requested_by.single()

    def to_text(self) -> str:
        return "The request titled '{title}'.".format(title=self.title)

    def to_json_connect(self, kg, json_dict) -> dict:
        out:dict= {}
        user:User = self.requested_by.single()
        
        if "user" not in json_dict:
            out["user"]=user.name
        if "email" not in json_dict:
            out["email"]=user.email

        return out

    def from_json_connect(self, kg, json_dict):
        # Ensure there is a requested_by relation
        if "user" in json_dict:
            user=kg.get_user_by_name(json_dict["user"])
            if user is None:
                email="unknown"
                if "email" in json_dict:
                    email = json_dict["email"]
                user=kg.add_user(json_dict["user"],email)
            self.save()
            self.requested_by.connect(user)

# Workflow Products | Recommendations and Summaries


class Recommends(nm.StructuredRel, OpenTldrMeta, OpenTldrEdge):
    '''
    Recommends edge connects a Recommendation node to the Content node that is being recommended.
    '''
    def to_text(self) -> str:
        return "This has {score} relevance for {content_type} '{content_title}'.".format(
            content_title=self.end_node().title,
            content_type= self.end_node().type,
            score=self.start_node.score_to_text())


class RelatesTo(nm.StructuredRel, OpenTldrMeta, OpenTldrEdge):
    '''
    RelatesTo edge connects a Recommendation node to the Request on which it is based.
    '''
    def to_text(self) -> str:
        return "This has {score} relevance to the request '{request}'.".format(
            request= self.end_node().title,
            score=self.start_node().score_to_text())


class Recommendation(nm.StructuredNode, OpenTldrMeta, OpenTldrNode, Scored):
    '''
    Recommendation nodes represent the assertion that a Content node (pointed to by Recommends edge)
    is believed to be relevant to a Request node (pointed to by the RelatesTo edge). The score
    property (float 0.0 - 1.0) indicates how relevant and thus how strong the recommendation.
    '''
    recommends = nm.RelationshipTo(Content, 'RECOMMENDS', model=Recommends, cardinality=One)
    relates_to = nm.RelationshipTo(Request, 'RELATES_TO', model=RelatesTo, cardinality=One)

    def get_recommends(self) -> Content:
        return self.recommends.single()
    
    def get_relates_to(self) -> Request:
        return self.relates_to.single()

    def to_text(self) -> str:
        return "The {content_type} '{content_title}' has {score} relevance to the request '{request}'.".format(
            content_title=self.recommends.single().text,
            content_type= self.recommends.single().type,
            request= self.relates_to.single().title,
            score=self.score_to_text())


class Summarizes(nm.StructuredRel, OpenTldrMeta, OpenTldrEdge):
    '''
    Summerizes edges connect a Summary node to the Content node that they summarize.
    '''
    pass


class FocusOn(nm.StructuredRel, OpenTldrMeta, OpenTldrEdge, Uncertain):
    '''
    FocusOn edges connect a Summary node to the Recommendation node that may help inform what
    the most relevant parts of the Summary are with respect to that Recommendation / Request.
    '''
    pass


class Summary(nm.StructuredNode, OpenTldrMeta, OpenTldrNode):
    '''
    Summary node contains text (String property) that is a shortened (ideally tailored) version
    of the Content node connect with the "summaries" edge. The "focus_on" edge connects to the
    Recommendation that may inform how this summary is tailored.
    '''
    text = nm.StringProperty(required=True)
    summarizes = nm.RelationshipTo(Content, 'SUMMARIZES', model=Summarizes, cardinality=One)
    focus_on = nm.RelationshipTo(Recommendation, 'FOCUS_ON', model=FocusOn, cardinality=ZeroOrOne)

    def get_summarizes(self) -> Content:
        return self.summarizes.single()
    
    def get_focus_on(self) -> Recommendation:
        return self.focus_on.single()


# Workflow Products | TLDR


class Includes(nm.StructuredRel, OpenTldrMeta, OpenTldrEdge):
    '''
    Includes edges connect a TldrEntry node to a Summary node to indicate that the Summary is
    to be used in that part of the TLDR.
    '''
    pass


class BasedOn(nm.StructuredRel, OpenTldrMeta, OpenTldrEdge):
    '''
    BasedOn edges connect a TldrEntry node to a Recommendation node to indicate why that
    entry is being included in the TLDR and may be used for things like ordering the entries
    to show the most recommended first.
    '''
    pass


class TldrEntry(nm.StructuredNode, OpenTldrMeta, OpenTldrNode, Scored):
    '''
    TldrEntry is one record within a TLDR. In general, there is a one-to-one relationship
    between each TldrEntry, Recommendation, and Summary such that each record within a TLDR
    has a score for the Recommendation, a text blurb from the Summary, and a connection back
    to the original Content (from both the Recommendation and Summary).
    '''
    link = nm.StringProperty(required=True)
    title = nm.StringProperty(required=True)
    type = nm.StringProperty(required=False)
    summary=nm.StringProperty(required=True)

    includes = nm.RelationshipTo(Summary, 'INCLUDES', model=Includes, cardinality=ZeroOrOne)
    based_on = nm.RelationshipTo(Recommendation, 'BASED_ON', model=BasedOn, cardinality=ZeroOrOne)

    def get_includes(self) -> Summary:
        return self.includes.single()
    
    def get_based_on(self) -> Recommendation:
        return self.based_on.single()

    def to_text(self) -> str:
        return "The entry summarizes {content_type} '{content_title}' and {score} relevance to the request.".format(
            content_title=self.title,
            content_type= self.type,
            score=self.score_to_text())


class Contains(nm.StructuredRel, OpenTldrMeta, OpenTldrEdge, Scored):
    '''
    Connects a Tldr to each of the TldrEntry objects that makes it up. The assumption
    here is that the score property (float 0.0 - 1.0) on the edge provides a descending ordering for
    the entries connected by these links.
    '''
    pass


class ResponseTo(nm.StructuredRel, OpenTldrMeta, OpenTldrEdge):
    '''
    ResponseTo edges connect a Tldr to the Request for which it was generated.
    '''
    pass


class Tldr(nm.StructuredNode, OpenTldrMeta, OpenTldrNode):
    '''
    Tldr nodes represent an instance of the TLDR report for the date (in 'date' property of type
    datetime.date). It is connected to the Request by the ResponseTo edge and a set of TldrEntries
    by "Entries" edges.
    '''
    date = nm.DateProperty(required=True)

    contains = nm.RelationshipTo(TldrEntry, 'CONTAINS', model=Contains, cardinality=ZeroOrMore)
    response_to = nm.RelationshipTo(Request, 'RESPONSE_TO', model=ResponseTo, cardinality=One)

    def get_contains(self) -> list[TldrEntry]:
        return self.contains

    def get_response_to(self) -> Request:
        return self.response_to.single()

    def to_text(self) -> str:
        out:str = "The TLDR for request '{request}' on date {date} includes: ".format(
            request= self.response_to.single().title,
            date=self.date)
        for tldr_entry in self.includes:
            out += "\t{text}\n".format(text=tldr_entry.to_text())
        return out



# FEEDBACK


class FeedbackForEntry(nm.StructuredRel, OpenTldrMeta, OpenTldrEdge):
    '''
    Edges connect a Feedback to the TLDR Entry for which it was generated.
    '''
    pass #TODO: complete this methods


class FeedbackForRequest(nm.StructuredRel, OpenTldrMeta, OpenTldrEdge):
    '''
    Edges connect a Feedback to the Request for which it was generated.
    '''
    pass #TODO: complete this method


class Feedback(nm.StructuredNode, OpenTldrMeta, OpenTldrNode, Scored):
    '''
    Feedback relates events back to a Request node to TldrEntry node to provide a method for
    storing events (for example giving it stars or logging accessing the source article)
    
    Feedback score ("score" property is float with expected range of 0.0 - 1.0) indicating
    how relevant that entry is with respect to the request. These edges are generally created
    by the UI when the user rates an entry (e.g., clicking on a star rating).

    Feedback clicked ("clicked" property is a date) indicates when the last time the source
    link for the TLDR entry was clicked by the user.
    '''
    click_date = nm.DateProperty(required=False)

    about_entry = nm.RelationshipTo(TldrEntry, 'ABOUT', model=FeedbackForEntry, cardinality=One)
    from_request = nm.RelationshipTo(Request, 'FROM', model=FeedbackForRequest, cardinality=One)
    
    def to_text(self) -> str:   
        click_status = "not clicked"
        if self.click_date is not None:
            click_status= "clicked on {date}".format(date=self.click_date.strftime("%m/%d/%Y"))
        
        rate_status = "unrated"
        if self.score is not None:
            rate_status = "rated as {rate}".format(rate=self.score_to_text())

        return "The feedback is that entry '{entry}' was {rate_status} and {click_status}.".format(
                content = self.key_for_content.single().to_text(),
                rate_status = rate_status,
                click_status = click_status);
       

# EVALUATION 


class KeyForContent(nm.StructuredRel, OpenTldrMeta, OpenTldrEdge):
    '''
    Edges connect a EvaluationKey to a Content node.
    '''
    pass #TODO: complete this method


class KeyForRequest(nm.StructuredRel, OpenTldrMeta, OpenTldrEdge):
    '''
    Edges connect a EvaluationKey to a Request.
    '''
    pass #TODO: complete this method


class EvalKey(nm.StructuredNode, OpenTldrMeta, OpenTldrNode, Scored):
    '''
    EvalKey is an indication of the relationship between a Content (e.g., a news article) and a
    Request (i.e., what the user is interested in). This represents a "correct" tailoring pair.
    This also includes a score and text to help identify what was most relevant in the pair.
    '''
    text = nm.StringProperty(required=True)
    key_for_content = nm.RelationshipTo(Content, 'KEY_FOR_CONTENT', model=KeyForContent, cardinality=One)
    key_for_request = nm.RelationshipTo(Request, 'KEY_FOR_REQUEST', model=KeyForRequest, cardinality=One)

    def to_json_connect(self, kg, json_dict) -> dict:
        out:dict= {}
        if "request" not in json_dict:
            out["request"]=self.key_for_request.single().title
        if "content" not in json_dict:
            out["content"]=self.key_for_content.single().title
        return out

    def from_json_connect(self, kg, json_dict):
        
        request:Request = None
        if "request" in json_dict:
            request=kg.get_request_by_title(json_dict["request"])
            if request is None:
                raise ValueError("Request titled '{req}' was not found in KG, but is required for EvalKey.".format(req=json_dict["request"]))
        else:
            raise ValueError("EvalKey requires a 'request' entry.")
        
        content:Content = None
        if "content" in json_dict:
            content=kg.get_content_by_title(json_dict["content"])
            if content is None:
                raise ValueError("Content titled '{req}' was not found in KG, but is required for EvalKey.".format(req=json_dict["content"]))
        else:
            raise ValueError("EvalKey requires a 'request' entry.")
        
        self.save()
        self.key_for_request.connect(request)
        self.key_for_content.connect(content)


    def to_text(self) -> str:
        return "The relevance of the content '{content}' to the request '{request}' is {amount}, and can be tailored down to '{text}'.".format(
            content = self.key_for_content.single().title,
            request = self.key_for_request.single().title,
            text = self.text,
            amount = self.score_to_text())



import logging

#logging.basicConfig(format='OpenTLDR Logging: %(message)s', level=logging.WARN)

logging.basicConfig(format='%(name)-12s: %(levelname)-8s %(message)s', level=logging.WARN)

log=logging.getLogger("OpenTLDR")



from abc import ABC, abstractmethod 
import re
import json

from .KnowledgeGraph import KnowledgeGraph
from .log import log

class AbstractDataRepo:
    '''
    baseclass for all Data Repositories, includes shared methods and ingest ordering.
    '''
    # knowledgegraph for import/export
    kg:KnowledgeGraph=None

    # order content should be imported to ensure dependencies.
    import_order=[
            "ReferenceNode",
            "ReferenceEdge",
            "Source",
            "Content",
            "User",
            "Request",
            "Feedback",
            "EvalKey"
        ]
  
    @abstractmethod
    def importData(self) -> list[str]:
        pass

    @abstractmethod
    def describe(self) -> str:
        pass

    def __init__(self, kg:KnowledgeGraph):
        self.kg=kg

    @classmethod
    def parseAnnotatedText(cls,text:str ) -> tuple[dict, str]:
        '''
        Parses a string from an annotated text file, extracts the annotations, and
        returns a tuple (dict of metadata , string of non annotation text block).

        Annotations are commented property key : value pairs, such as:
            # uid = 1234\n
            This is a text block.\n
            # name = Chris\n
            Another line.\n
            And one at the end.\n

        Will return ({'uid': '1234', 'name':'Chris}, 'This is a text block.\nAnother line.\nAnd one at the end')    
        '''
        # Identifies a property key
        property_regex=re.compile(r'^#\s*[a-z_]+\s*\:\s*')
        
        content_text=""
        properties={}

        for line in text.splitlines():
            match= property_regex.search(line)
            if match:
                # This line is an annotation
                key = re.sub("[#:\s]","",match.group()).lower()
                value = ''.join(re.split(property_regex,line)).strip()
                properties[key]=value
            else:
                # This line is a text block
                content_text='\n'.join([content_text,line])

        return (properties, content_text.strip())
    
    @classmethod
    def inferDomainClass(cls,full_path:str,annotations:dict) -> str:
        if 'class' in annotations:
            return annotations["class"]
        log.debug("Class was not specified in annotations, attempting to infer from path ({path}).".format(path=full_path))
        path = full_path.lower()
        if 'reference' in path:
            return "ReferenceNode"
        if 'active' in path or 'content' in path or 'article' in path:
            return "Content"
        if 'request' in path or 'query' in path or 'user' in path:
            if 'text' in annotations:
                return "Request"
            if 'name' in annotations:
                return "User"
            # TODO: Nothing is returned if no text and no name
        if 'eval' in path or 'evalkey' in path or 'evaluation' in path:
            return "EvalKey"
        if 'feedback' in path:
            return "Feedback"

    def clean_up_nodes(self) -> list[str]:
        out:list[str]=[]
        for tag in self.import_order:
            log.info("Cleaning up {tag}:".format(tag=tag))
            node_list = self.kg.get_all_nodes_by_tag(tag)
            if node_list is None:
                continue
            for node in node_list:
                node.save() # Updates properties, including uid
                out.append(node.uid)
        log.info("Cleaned {n} nodes.".format(n=len(out)))
        return out

    def exportData(self, path:str):
        pass


from .DataRepo import AbstractDataRepo
from .KnowledgeGraph import KnowledgeGraph
from .Domain import *
from .log import log

import os
import json

class DataRepoJsonFiles(AbstractDataRepo):
    '''
    Reads a single .json text file from the path and populates the KG using it.
    This is a great way to export checkpoints for a KG and re-import them later.
    '''

    def __init__(self, kg:KnowledgeGraph, filename:str):
        self.kg=kg
        self.filename = filename
        self.verify()


    def verify(self):

        path = self.filename

        # ensure path points to a readable file
        if not os.path.exists(path):
            raise ValueError ("Cannot reach path: {path}".format(path=path))
        
        if os.path.isdir(path):
            raise ValueError ("Path ({path}) is a directory, looking for a .json file.".format(path=path))

        if not os.access(path,os.R_OK):
            raise ValueError("Cannot read file at {path}".format(path=path))
        
        with open(path) as f:
            self.data = json.load(f)


    def describe(self) -> str:
        return "JSON File (path = '{path}')".format(path=self.ingest_path)  


    def importAll(self) -> dict:
        out:dict={}
        out.update(self.importReferenceData())
        out.update(self.importActiveData())
        out.update(self.importRequestData())
        out.update(self.importFeedbackData())
        out.update(self.importEvaluationData())
        return out
    

    def importReferenceData(self) -> dict:
        log.info("Importing Reference Data from JSON File...")
        node_look_up:dict={}

        # Import Reference Nodes
        node_uids:list[str] = []
        for ref_node_json in self.data['reference']['nodes']:
            ref_node = ReferenceNode.from_json(ref_node_json)
            ref_node.save()
            node_uids.append(ref_node.uid)
            node_look_up[ref_node.uid]=ref_node
        log.info("Imported {c} Reference Nodes.".format(len(node_uids)))

        # Import Reference Edges
        edge_uids:list[str] = []
        for ref_edge_json in self.data['reference']['edges']:
            ref_edge = ReferenceEdge.from_json(ref_edge_json)
            # TODO: set the to/from uids into neomodel object
            ref_edge.save()
            edge_uids.append(ref_edge.uid)
        log.info("Imported {c} Reference Edges.".format(len(edge_uids)))

        return {"ReferenceNodes":node_uids, "ReferenceEdges":edge_uids}
    

    def importRequestData(self) -> list[str]:
        log.info("Importing Requests from File System...")

        path = self.ingest_path
        alt_path = os.path.join(self.ingest_path,"requests")
        if os.path.exists(alt_path) and os.path.isdir(alt_path) and os.access(alt_path,os.R_OK):
            log.debug("Found data type directory in specified repo path.")
            path=alt_path

        default_source_name = "File System Directory:{path}".format(path=path)      # TODO: Never used.  Delete.
        list_of_uids = []

        # read in each text file from directory
        for filename in os.listdir(path):

            # ignore anything but files that end in .txt
            if os.path.splitext(filename)[1] != ".txt" or not os.path.isfile(os.path.join(path, filename)):
                continue

            full_path=os.path.join(path, filename)
            with open(full_path) as f:
                text = f.read()
                request=self.importTextRequest(kg=self.kg, text=text)       # TODO: importTextRequest never defined
                list_of_uids.append(request.uid)
                log.info("Imported: {data}".format(data=request.to_text()))

        return list_of_uids

    def importActiveData(self) -> list[str]:
        log.info("Importing Active Data from File System...")
        path = self.ingest_path
        alt_path = os.path.join(self.ingest_path,"active")
        if os.path.exists(alt_path) and os.path.isdir(alt_path) and os.access(alt_path,os.R_OK):
            log.debug("Found data type directory in specified repo path.")
            path=alt_path
        default_source_name = "File System Directory:{path}".format(path=path)
        list_of_uids = []

        # read in each text file from directory
        for filename in os.listdir(path):
            full_path=os.path.join(path, filename)

            # ignore anything but files that end in .txt
            if os.path.splitext(filename)[1] != ".txt" or not os.path.isfile(full_path):
                continue

            with open(full_path) as f:
                text = f.read()
                url="file://{path}".format(path=os.path.abspath(full_path))
                content=self.importTextContent(kg=self.kg, text=text, url=url, default_source_name=default_source_name)  # TODO: importTextRequest never defined
                list_of_uids.append(content.uid)
                log.info("Imported: {data}".format(data=content.to_text()))

        return list_of_uids
    
    def importFeedbackData(self) -> list[str]:
        message:str = "Importing MongoDB Feedback Data is not implemented yet."
        log.error(message)
        raise NotImplementedError(message)

    def importEvaluationData(self) -> list[str]:
        message:str = "Importing MongoDB Evaluation Data is not implemented yet."
        log.error(message)
        raise NotImplementedError(message)


import os
from dotenv import load_dotenv

import datetime

from neo4j import GraphDatabase
import neomodel as nm

from opentldr.Domain import *
from opentldr.ContentEnrichment import *
from .log import log

load_dotenv()
load_dotenv("{d}/.env".format(d=os.getcwd()))
load_dotenv("{d}/../.env".format(d=os.getcwd()))

def _getenv(variable:str , default:str):
        '''
        _getenv(variable, default) attempts to resolve the value of the environment variable specified.
        If it cannot find the variable in the OS or .env (from dotenv package) it will fail over to the
        provided default value while giving a warning to the logging system.
        '''
        value:str = os.getenv(variable)
        if value == None:
            log.warning("No environment variable '"+variable+"' so defaulting to '"+default+"'.")
            return default
        return value

class KnowledgeGraph:
    '''
    KnowledgeGraph is the API for OpenTLDR's interface to Neo4j graph database. It attempts to make the
    driver connection using OS, .env, and then default values for connections. This API maps neo4j
    nodes/edges to types specified in the Domain module (using Neomodel OGM and some custom cypher).

    In order to be successful, you will need to have a neo4j server running, reachable on the network,
    and with environment variables (either in OS or .env files) similar to these:

        NEO4J_CONNECTION='neo4j://localhost:7687'
        NEO4J_USERNAME=neo4jUser
        NEO4J_PASSWORD=neo4jPassword
        NEO4J_DATABASE=neo4j

    This class can be used in several different ways, including:
        (1) import opentldr.KnowledgeGraph ... kg=KnowledgeGraph() ... (use kg here) ... kg.close() 
        (2) with opentldr.KnowledgeGraph as kg: ... (use kg here, and then it will autoclose)

    '''

    # -----------------
    # Setup Methods
    # -----------------

    # Default values come from first of: (1) .env file, (2) os environment, or (3) these hardcoded values
    connection=_getenv('NEO4J_CONNECTION','neo4j://localhost:7687')
    user= _getenv('NEO4J_USERNAME','neo4j')
    password= _getenv('NEO4J_PASSWORD','neo4jPassword')
    database= _getenv('NEO4J_DATABASE','neo4j')
    driver = GraphDatabase.driver(connection, auth=(user,password))
    nm.config.DRIVER = driver

    def __init__(self, neo4j_driver=None, connection=None, user=None, password=None, database=None):
        if connection is not None:
            self.connection=connection
        if user is not None:
            self.user=user
        if password is not None:
            self.password=password
        if database is not None:
            self.database=database
         
        if neo4j_driver is None:        
            log.info("Connecting to Neo4J using environment specification.")
            log.info("\tconnection:\t{v}".format(v=self.connection))
            log.info("\tuser:\t\t{v}".format(v=self.user))
            #log.info("\tpassword:\t{v}".format(v=self.password))           # don't display sensitive information
            log.info("\tdatabase:\t{v}".format(v=self.database))
            self.connect()
        else:
            log.info("Using Neo4j driver passed into initialization.")
            self.connect(neo4j_driver)

    def __del__(self):
        if self.driver is not None:
            self.driver.close()

    def __enter__(self):
        return self
    
    def __exit__(self,ctx_type, ctx_value, ctx_traceback):
        if self.driver is not None:
            self.driver.close()

    # -----------------
    # General Utility Methods
    # -----------------

    def neo4j_query(self, cql):
        '''
        Performs a cypher call to the knowledge base and returns results in the standard neo4j driver format.
        Use this if you are following neo4j documentation, otherwise use cypher_query().
        '''
        if self.driver == None:
            self.connect()

        try:
            records, summary, keys = self.driver.execute_query(cql, database_=self.database,)
            log.debug("The query `{query}` returned {records_count} records in {time} ms.".format(
                query=summary.query, records_count=len(records),
                time=summary.result_available_after))
            return records, summary
        except Exception as e:
            log.error(e)


    def neomodel_query(self,cypher_query, params: dict={}):
        '''
        Performs a cypher query in the standard neomodel method, which will return a list of lists of neomodel.
        Use this if you are following Neomodel documentation, otherwise use cypher_query().
        '''
        if self.driver == None:
            self.connect()
        
        return nm.db.cypher_query(cypher_query, resolve_objects=True, params=params)


    def connect(self, driver = None):
        '''
        Ensures that a connection to Neo4j has been established and is functioning.
        '''
        try:
            if driver is None:
                if self.driver is None:
                    driver = GraphDatabase.driver(self.connection, auth=(self.user,self.password))
                else:
                    driver = self.driver
    
            log.info("Testing Neo4j Connectivity...")
            driver.verify_connectivity()
            
            if driver.verify_connectivity:
                log.info("Successfully connected to Neo4J server for KnowledgeGraph.")
            else:
                log.error("Unable to connect to Neo4J for KnowledgeGraph. Check that Neo4j server is running.")
                return False

            log.info("Testing Neo4j Authentication...")
            driver.verify_authentication()
            
            if driver.verify_authentication:
                log.info("Successfully logged into Neo4J server for KnowledgeGraph.")
            else:
                log.error("Unable to log in to Neo4J for KnowledgeGraph. Check that your user and password are correct.")
                return False
            
            self.driver = driver
            nm.config.DRIVER = self.driver
            nm.db.set_connection(driver=self.driver)

            return
        
        except Exception as e:
            log.error(e)
            log.error("There was an issue connecting to Neo4J server, please verify that it is running and configured correctly.")            
            raise e
    

    def getNeo4jDriver(self):
        '''
        Returns the Neo4j driver class being used (usually neo4j.GraphDatabase.driver) so that direct neo4j calls can be
        made or it can be passed into other libraries.
        '''
        if self.driver is None:
            self.connect()
        return self.driver

    def close(self):
        '''
        Shuts down the current connection to neo4j. This should be done at the end of each program / notebook.
        If you use the "with statement" context control, this will be called automatically when the code block ends.
        '''
        if self.driver is not None:
            self.driver.close()
        self.driver=None

    def _ensure_saved(self,node:OpenTldrMeta):
        '''
        OpenTLDR objects all inherit from the OpenTldrMeta mixin class. They will be assigned a "uid" when they are saved.
        So, we can use this to determine if an object has been saved or not and respond appropriately.
        '''
        if node.uid is None:
            log.warning("Nodes must be persisted with node.save() before being connected. Forcing this now.")
            return node.save()
        return node

    def get_all_node_uids_by_tag(self, tag:str) -> list[str]:
        '''
        Returns the list of uids for the tag specified as a string.
        '''
        return self.cypher_query("MATCH (n:{tag}) RETURN n.uid".format(tag=tag))

    def get_all_nodes_by_tag(self, tag:str):
        '''
        Returns the list of uids for the tag specified as a string.
        '''
        return self.cypher_query("MATCH (n:{tag}) RETURN n".format(tag=tag),"n")

    # -----------------
    # KnowedgeGraph API
    # -----------------

    def cypher_query(self,cypher_query, extract_variable=None, params: dict={}) -> list:
        '''
        Performs a cypher query and returns a single simple list of Neomodel objects based on the variable indicated (or first).
        Use this for most OpenTLDR queries because it is much simpler to process for iterating on the results.
        For example: cypher_query("MATCH (a)-[b]->(c) RETURN a,b,c","a") only returns a list of only the "a" results.
        So, cypher_query("MATCH (a)-[b]->(c) RETURN a","a") would give the same result list by dropping the "b" and "c" results.
        Reduce the risk of Cypher injection attacks by passing Cypher parameters in params.
        '''
        if self.driver is None:
            log.debug("Cypher Query made before driver created, creating connection driver.")
            self.connect()
        
        results, meta = nm.db.cypher_query(cypher_query, resolve_objects=True, params=params)
        
        index=0
        if extract_variable is not None:
            index=meta.index(extract_variable)
        else:
            if len(meta)>1:
                log.warning("cypher_query is defaulting to return the first element of multiple return values in results.")      

        return [item[index] for item in results]

    def cypher_query_one(self,cypher_query,extract_variable=None, params: dict={}):
        '''
        Performs a query that is intended to LIMIT the return values to one single Neomodel object. This simplifies the processing
        of the query results to consistently only get a single object for the extract_variable, it is undefined which unless you sort them.
        For example: cypher_query_one("MATCH (a)-[b]->(c) RETURN a,b,c","a") returns the FIRST object of result "a".
        So, cypher_query_one("MATCH (a)-[b]->(c) RETURN a LIMIT 1","a") gives the same result (faster).
        Reduce the risk of Cypher injection attacks by passing Cypher parameters in params.
       '''
        results = self.cypher_query(cypher_query,extract_variable, params=params)

        # Limits the output to the first actual object (i.e., not a list)        
        if results is not None and len(results)>0:
            return results[0]
        else:
            return None  
        
    def cypher_import(self,filename):
        '''
        Processes a text file containing cypher commands and executes them on the KnowledgeGraph.
        '''
        count=0
        with open(filename) as fp:
            lines = fp.readlines()
            for line in lines:
                if len(line.strip()) == 0:
                    continue
                with self.driver.session() as session:
                    session.run(line)
                count=count+1
        log.info("Imported {count} lines of Cypher from {filename}".format(count=count, filename=filename))

    def delete_all(self):
        '''
        Clears the neo4j graph database.
        '''
        self.cypher_query("MATCH(x) DETACH DELETE x")

    def shortest_path(self,start:OpenTldrMeta,end:OpenTldrMeta):
        '''
        This is a special cypher command to run neo4j's shortest path algorithm and return the path (list of edges) found.
        We remove EvalKey, Recommendation, Tldr, Source, and User nodes to avoid irrelevant paths that circumvent Reference Data.
        '''
        results,meta = self.neomodel_query("""
            MATCH path=shortestPath((s)-[*..10]-(e))
            WHERE s.uid='{start_id}'
            AND e.uid='{end_id}'                         
            AND NONE(n IN nodes(path) WHERE 'Recommendation' IN LABELS(n))
            AND NONE(n IN nodes(path) WHERE 'Tldr' IN LABELS(n))
            AND NONE(n IN nodes(path) WHERE 'Source' IN LABELS(n))
            AND NONE(n IN nodes(path) WHERE 'User' IN LABELS(n))
            AND NONE(n IN nodes(path) WHERE 'EvalKey' IN LABELS(n))
            RETURN path
            """.format(start_id=start.uid, end_id=end.uid))    
                     
        if len(results) == 0:
            return None
        else:
            return results[0][meta.index("path")]

    def get_by_uid(self, uid:str):
        return self.cypher_query_one("MATCH (n) WHERE n.uid=$uid RETURN n", params={"uid": uid})

    # -----------------
    # Reference Data API
    # -----------------

    ## Reference Nodes in KG are labeled as ReferenceNode

    def add_reference_node(self, text:str, type:str, hypothesized:bool=False) -> ReferenceNode:
        '''
        Creates a new ReferenceNode object in KG.
        '''
        return ReferenceNode(text=text, type=type, hypothesized=hypothesized).save()
    
    def get_reference_node_by_uid(self,uid:str) -> ReferenceNode:
        '''
        Returns a single ReferenceNode as indicated by the uid or None if it does not exist in KG.
        '''
        return ReferenceNode.nodes.get_or_none(uid=uid);

    def get_all_reference_nodes(self) -> list[ReferenceNode]:
        '''
        Returns a list of all the ReferenceNodes in the KG.
        '''
        return self.cypher_query("MATCH (n:ReferenceNode) RETURN n","n")

    def update_reference_node(self, node:ReferenceNode) -> ReferenceNode:
        '''
        Updates the KG with changes to a ReferenceNode properties
        '''
        return node.save()
    
    def delete_reference_node(self, node:ReferenceNode):
        '''
        Removes the passed ReferenceNode and the edges connected to them.
        '''
        self.cypher_query("MATCH (n:ReferenceNode) WHERE n.uid=$uid DETACH DELETE n", params={"uid": node.uid})
   
    def delete_all_reference_nodes(self):
        '''
        Removes all of the ReferenceNode and the edges connected to them.
        '''
        self.cypher_query("MATCH (n:ReferenceNode) DETACH DELETE n")

    def get_reference_node_by_uid(self,uid:str) -> ReferenceNode:
        '''
        Returns a single ReferenceNode as indicated by the uid or None if it does not exist in KG.
        '''
        return ReferenceNode.nodes.get_or_none(uid=uid);

    def get_all_reference_nodes(self) -> list[ReferenceNode]:
        '''
        Returns a list of all the ReferenceNodes in the KG.
        '''
        return self.cypher_query("MATCH (n:ReferenceNode) RETURN n","n")

      
  ## Reference Edges - in KG these are labeled as REFERENCE_EDGE but Neomodel classes are ReferenceEdge


    def add_reference_edge(self, from_node:ReferenceNode, to_node:ReferenceNode, text:str, type:str, hypothesized:bool=False):
        '''
        Adds a ReferenceEdge between two ReferenceNodes. Note these edges are only between ReferenceNodes.
        '''
        from_node:ReferenceNode=self._ensure_saved(from_node)
        to_node:ReferenceNode=self._ensure_saved(to_node)
        edge=from_node.edges.connect(to_node,{'type':type, 'text':text, 'hypothesized':hypothesized})
        return edge.save()
    
    def get_reference_edge_by_uid(self,uid:str) -> ReferenceEdge:
        '''
        Returns a single ReferenceEdge as indicated by the uid or None if it does not exist in KG.
        '''
        self.cypher_query_one("MATCH (a)-[r:REFERENCE_EDGE]->(b) WHERE r.uid=$uid RETURN r", params={"uid": uid})

    def get_all_reference_edges(self) -> list[ReferenceEdge]:
        '''
        Returns a list of all of the ReferenceEdges in the KG.
        '''
        return self.cypher_query("MATCH (a)-[n:REFERENCE_EDGE]->(b) RETURN n","n")

    def update_reference_edge(self, edge:ReferenceEdge) -> ReferenceEdge:
        '''
        Updates the KG with changes to a ReferenceEdge's properties
        '''
        return edge.save()

    def delete_reference_edge(self, edge:ReferenceEdge):
        '''
        Removes the ReferenceEdge.
        '''
        self.cypher_query("MATCH (a)-[r:REFERENCE_EDGE]->(b) WHERE r.uid=$uid DETACH DELETE r", params={"uid": edge.uid})

    def delete_all_reference_edges(self):
        '''
        Removes all of the ReferenceEdges from KG.
        '''
        self.cypher_query("MATCH (a)-[r:REFERENCE_EDGE]->(b) DETACH DELETE r")

    # =================
    # Active Data
    # =================

    # -----------------
    # Source Nodes - refer to the origin of content
    # -----------------

    def add_source(self, name:str) -> Source:
        '''
        Creates new Source node, or if one by the same name already exists it reuses it.
        '''
        source = Source.nodes.get_or_none(name=name)
        if source is None:
            source= Source(name=name).save()
        else:
            log.warning("Found existing Source node for name {name}, returning pre-existing node.".format(name=name))
        return source

    def get_source_by_uid(self, uid: str) -> Source:
        '''
        Gets an existing Source node by uid or None if it does not exist
        '''
        return Source.nodes.get_or_none(uid=uid)

    def get_source_by_name(self, name: str) -> Source:
        '''
        Gets an existing Source node by name or None if it does not exist
        '''
        return Source.nodes.get_or_none(name=name)

    def get_all_sources(self) -> list[Source]:
        return Source.nodes

    def update_source(self, node:Source) -> Source:
        '''
        Updates the KG with changes to a Source's properties.
        '''
        return node.save()
    
    def delete_source(self, source:Source):
        '''
        Removes the passed Source node, its content nodes, and the edges connected to it.
        '''
        for c in self.get_content_by_source(source):
            self.delete_content(c)
        return self.delete_source_by_uid(source.uid)
   
    def delete_source_by_uid(self, uid:str):
        return self.cypher_query("""
            MATCH (s:Source) WHERE s.uid=$uid
            DETACH DELETE s""",
            params={"uid": uid})

    def delete_all_sources(self):
        '''
        Removes all of the Sources, their content, and the edges connected to them.
        '''
        for s in self.get_all_sources():
            self.delete_source(s)

    # -----------------
    # Content Nodes - contain article content and link to their Source
    # -----------------

    def add_content(self, source:Source, type:str, url:str, date:datetime.date, title:str, text:str) -> Content:
        content = Content.nodes.get_or_none(url=url)
        if content is None:
            content=Content(url=url,date=date,title=title,text=text, type=type).save()
            content.is_from.connect(source)
        else:
            log.warning("Found existing Content node for url {url}, returning pre-existing node.".format(url=url))
        return content

    def get_content_by_uid(self, uid:str) -> Content:
        content = Content.nodes.get_or_none(uid=uid)
        return content

    def get_content_by_title(self, title: str) -> Request:
        return self.cypher_query_one("""
            MATCH (c:Content) WHERE c.title=$title 
            RETURN c """,
            "c", params={"title": title})

    def get_content_by_url(self, url:str) -> Content:
        content = Content.nodes.get_or_none(url=url)
        return content

    def get_content_by_date(self,date:datetime.date) -> list[Content]:
        content = Content.nodes.get_or_none(date=date)
        return content

    def get_content_by_source(self,source:Source) -> list[Content]:
        return self.get_content_by_source_uid(source.uid)

    def get_content_by_source_uid(self, uid: str) -> list[Content]:
        return self.cypher_query("""
            MATCH (s:Source) WHERE s.uid=$uid
            MATCH (c:Content)
            MATCH (c)-[y:IS_FROM]->(s)
            RETURN c """,
            "c", params={"uid": uid})

    def get_content_by_recommendation(self,recommendation:Recommendation) -> list[Content]:
        return recommendation.get_recommends()

    def get_all_content(self) -> list[Content]:
        content = Content.nodes
        return content

    def delete_content(self, content:Content):
        for e in self.get_entities_by_content(content):
            self.delete_entity(e)
        for k in self.get_evalkeys_by_content(content):
            self.delete_evalkey(k)
        return self.delete_content_by_uid(content.uid)

    def delete_content_by_uid(self, uid:str):
        return self.cypher_query('''
            MATCH (c:Content) WHERE c.uid=$uid
            DETACH DELETE c ''',
            params={"uid": uid})

    def delete_all_content(self):
        for c in self.get_all_content():
            self.delete_content(c)


    # -----------------
    # Active Data / Entity
    # -----------------

    def add_entity(self, node:Content|Request, text:str, type:str ) -> Entity:
        entity=Entity(text=text,type=type).save()
        entity.mentioned_in.connect(node)
        return entity
    
    def add_refers_to_edge(self, entity:Entity, reference:ReferenceNode, confidence=1.0 ):
        entity.refers_to.connect(reference, {'confidence': confidence })

    def get_entity_by_uid(self, uid:str) -> Entity:
        entities = Entity.nodes.get_or_none(uid=uid)
        return entities
     
    def get_entity_by_url(self, url:str) -> Entity:
        entities = Entity.nodes.get_or_none(url=url)
        return entities
    
    def get_entities_by_content(self,content:Content) -> list[Entity]:
        return self.cypher_query("""
            MATCH (c:Content) WHERE c.uid=$content_uid
            MATCH (e:Entity)
            MATCH (e)-[y:MENTIONED_IN]->(c)
            RETURN e """,
            "e", params={"content_uid": content.uid})
    
    def get_entities_by_content_uid(self,uid:str) -> list[Entity]:
        return self.cypher_query("""
            MATCH (c:Content) WHERE c.uid=$content_uid
            MATCH (e:Entity)
            MATCH (e)-[y:MENTIONED_IN]->(c)
            RETURN e """,
            "e", params={"content_uid": uid})

    def get_entities_by_request(self,request:Request) -> list[Entity]:
        return self.cypher_query("""
            MATCH (r:Request) WHERE r.uid=$request_uid
            MATCH (e:Entity)
            MATCH (e)-[y:MENTIONED_IN]->(r)
            RETURN e """,
            "e", params={"request_uid": request.uid})

    def get_all_entities(self) -> list[Entity]:
        entities = Entity.nodes
        return entities

    def delete_entity(self, node:Entity):
        return self.delete_entity_by_uid(node.uid)

    def delete_entity_by_uid(self, uid:str):
        return self.cypher_query("""
            MATCH (e:Entity) WHERE e.uid=$uid
            DETACH DELETE e """,
            params={"uid": uid})

    def delete_all_entities(self):
        return self.cypher_query("""
            MATCH (e:Entity)
            DETACH DELETE e """)

    # -----------------
    # Requests API
    # -----------------

    # -----------------
    # Users
    # -----------------


    def add_user(self, name:str, email:str) -> User:
        user = User.nodes.get_or_none(name=name)
        if user is None:
            user=User(name=name, email=email).save()
        else:
            log.warning("Found existing User node for name {name}, returning pre-existing node.".
                      format(name=name))
        return user      

    def get_user_by_uid(self,uid:str) -> User:
        user = User.nodes.get_or_none(uid=uid)
        return user

    def get_user_by_name(self,name:str) -> User:
        user = User.nodes.get_or_none(name=name)
        return user
    
    def get_all_users(self) -> list[User]:
        users = User.nodes
        return users

    def delete_user(self, user:User):
        for r in self.get_requests_by_user(user):
            self.delete_request(r)
        return self.delete_user_by_uid(user.uid)

    def delete_user_by_uid(self, uid:str):
        return self.cypher_query("""
            MATCH (u:User) WHERE u.uid=$uid
            DETACH DELETE u""",
            params={"uid": uid})

    def delete_all_users(self):
        for u in self.get_all_users():
            self.delete_user(u)


    # -----------------
    # Requests
    # -----------------

    def add_request(self, title:str, text:str, user:User) -> Request:
        request = Request.nodes.get_or_none(title=title)
        if request is None or request.requested_by.single().uid != user.uid:
            request = Request(title=title, text=text).save()
            request.requested_by.connect(user)
        else:
            log.warning("Found existing Request node for {title}, returning pre-existing node.".format(title=title))
        return request

    def get_request_by_uid(self, uid:str) -> Request:
        request = Request.nodes.get_or_none(uid=uid)
        return request

    def get_requests_by_user(self, user: User) -> list[Request]:
        return self.get_requests_by_user_uid(user.uid)

    def get_requests_by_user_uid(self, uid: str) -> list[Request]:
        return self.cypher_query("""
            MATCH (u:User) WHERE u.uid=$user_id
            MATCH (q:Request)
            MATCH (q)-[y:REQUESTED_BY]->(u)
            RETURN q """,
            "q", params={"user_id": uid})

    def get_request_by_title(self,title:str) -> Request:
       return self.cypher_query_one("""
            MATCH (q:Request) WHERE q.title=$title
            RETURN q """,
            "q", params={"title": title})

    def get_all_requests(self) -> list[Request]:
        requests = Request.nodes
        return requests

    def delete_request(self,request:Request):
        for e in self.get_entities_by_request(request):
            self.delete_entity(e)
        for t in self.get_tldr_by_request(request):
            self.delete_tldr(t)
        for k in self.get_evalkeys_by_request(request):
            self.delete_evalkey(k)   
        return self.delete_request_by_uid(request.uid)
    
    def delete_request_by_uid(self,uid:str):
        return self.cypher_query('''
            MATCH (r:Request) WHERE r.uid=$uid
            DETACH DELETE r ''',
            params={"uid": uid})
    
    def delete_all_requests(self):
        for r in self.get_all_requests():
            self.delete_request(r)
   
    def get_requests_by_tldr_entry(self,entry:TldrEntry) -> Request: 
        return self.cypher_query_one("""
            MATCH (e:TldrEntry) WHERE e.uid=$uid
            MATCH (t:Tldr)
            MATCH (q:Request)
            MATCH (e)<-[x:CONTAINS]-(t)-[y:RESPONSE_TO]->(q)
            RETURN q ""","q", params={'uid': entry.uid})
    

    # -----------------
    # Workflow Products API
    # -----------------

    # -----------------
    # Recommendations
    # -----------------

    def add_recommendation(self, score:float, content:Content, request:Request) -> Recommendation:
        if content.uid is None or request.uid is None:
            log.error("Adding Recommendation requires an existing and saved Content and Request node. Attempting to save them...")
            content.save()
            request.save()
        
        recommendation = self.get_recommendation(content,request)
        if recommendation is None:
            recommendation= Recommendation(score=score).save()
            recommendation.recommends.connect(content)
            recommendation.relates_to.connect(request)
        else:
            log.warning("Found existing Recommendation node between content and request, returning pre-existing node.")
        return recommendation
    
    def get_recommendation(self, content:Content, request:Request) -> Recommendation:
        return self.cypher_query_one("""
            MATCH (q:Request) WHERE q.uid=$request_id
            MATCH (a:Content) WHERE a.uid=$content_id
            MATCH (r:Recommendation)
            MATCH (r)-[x:RECOMMENDS]->(a)
            MATCH (r)-[y:RELATES_TO]->(q)
            RETURN r LIMIT 1 """,
            "r", params={
                "request_id": request.uid,
                "content_id": content.uid})

    def get_recommendation_by_id(self, uid:str) -> Recommendation:
        recommendation = Recommendation.nodes.get_or_none(uid=uid)
        return recommendation

    def get_recommendations_by_request(self,request:Request) -> list[Recommendation]:
        return self.cypher_query("""
            MATCH (q:Request) WHERE q.uid=$request_id
            MATCH (r:Recommendation)
            MATCH (r)-[y:RELATES_TO]->(q)
            RETURN r ORDER BY r.score DESC """,
            "r", params={"request_id": request.uid})

    def get_all_recommendations(self) -> list[Recommendation]:
        recommendations = Recommendation.nodes
        return recommendations

    def delete_recommendation(self, recommendation:Recommendation):
        recommendation.delete()
    
    def delete_all_recommendations(self): 
        self.cypher_query("MATCH (r:Recommendation) DETACH DELETE (r)")
         
    # -----------------
    # Summaries
    # -----------------

    def add_summary(self, text:str, content:Content, recommendation:Recommendation=None) -> Summary:
        summary = self.get_summary(content,recommendation)
        if summary is None:
            summary= Summary(text=text).save()
            summary.summarizes.connect(content)
            if recommendation is not None:
                summary.focus_on.connect(recommendation)
        else:
            log.warning("Found existing Summary node, returning pre-existing node.")
        return summary
    
    def get_summary(self, content:Content, recommendation:Recommendation=None) -> Summary:
        # Summaries without recommendations should explictly NOT have recommendations associated with them
        if recommendation is None:
            return self.cypher_query_one("""
                MATCH (a:Content) WHERE a.uid=$content_id
                MATCH (s:Summary)
                MATCH (s)-[x:SUMMARIZES]->(a)
                WHERE NOT (s)-[:FOCUS_ON]->()
                RETURN s """,
                "s", params={"content_id": content.uid})
        else:
            return self.cypher_query_one("""
                MATCH (r:Recommendation) WHERE r.uid=$recommendation_id
                MATCH (a:Content) WHERE a.uid=$content_id
                MATCH (s:Summary)
                MATCH (s)-[x:SUMMARIZES]->(a)
                MATCH (s)-[y:FOCUS_ON]->(r)
                RETURN s """,
                "s", params={
                    "recommendation_id": recommendation.uid,
                    "content_id": content.uid})
        
    def get_summary_by_id(self, uid:str) -> Summary:
        summary = Summary.nodes.get_or_none(uid=uid)
        return summary

    def get_summaries_by_content(self, content:Content) -> list[Summary]:
        return self.cypher_query("""
            MATCH (a:Content) WHERE a.uid=$content_id
            MATCH (s:Summary)
            MATCH (s)-[x:SUMMARIZES]->(a)
            RETURN s """,
            params={"content_id": content.uid})

    def get_summaries_by_recommendation(self, recommendation:Recommendation) -> list[Summary]:      
        return self.cypher_query("""
            MATCH (r:Recommendation) WHERE r.uid=$recommendation_id
            MATCH (s:Summary)
            MATCH (s)-[x:FOCUS_ON]->(r)
            RETURN s """,
            "s", params={"recommendation_id": recommendation.uid})

    def get_all_summary(self) -> list[Summary]:
        summary = Summary.nodes
        return summary
    
    def delete_summary(self, summary:Summary):
        summary.delete()
    
    def delete_all_summaries(self): 
        self.cypher_query("MATCH (s:Summary) DETACH DELETE (s)")
    
    # -----------------
    # Tldr
    # -----------------

    def add_tldr(self, request: Request, date: datetime.date) -> Tldr:
        tldr = self.get_tldr(request, date)
        if tldr is None:
            tldr = Tldr(date=date).save()
            tldr.response_to.connect(request)
        else:
            log.warning("Found existing Tldr node for request on given date, returning pre-existing node.")
        return tldr

    def get_tldr(self, request: Request, date: datetime.date) -> Tldr:
        return self.get_tldr_by_uid_and_date(request.uid, date)

    def get_tldr_by_uid(self, uid: str) -> Tldr:
        tldr = Tldr.nodes.get_or_none(uid=uid)
        return tldr

    def get_tldrs_by_date(self, date: datetime.date):
        results = self.cypher_query("""
            MATCH (t:Tldr) WHERE t.date=$date
            RETURN t """,
            "t", params={"date": date})
        if results is None:
            return None
        else:
            return results[0]

    def get_tldrs_before_date(self, date: datetime.date):
        return self.cypher_query("""
            MATCH (t:Tldr) WHERE t.date<$date
            RETURN t """,
            "t", params={"date": date})

    def get_tldr_by_request(self, request: Request) -> list[Tldr]:
        return self.get_tldr_by_request_uid(request.uid)

    def get_tldr_by_request_uid(self, request_uid: str) -> list[Tldr]:
        return self.cypher_query("""
            MATCH (q:Request) WHERE q.uid=$request_uid
            MATCH (t:Tldr)
            MATCH (t)-[y:RESPONSE_TO]->(q)
            RETURN t """,
            "t", params={"request_uid": request_uid})

    def get_tldr_by_uid_and_date(self, request_uid: str, date: datetime.date) -> Tldr:
        return self.cypher_query_one("""
            MATCH (q:Request) WHERE q.uid=$request_uid
            MATCH (t:Tldr) where t.date=$date
            MATCH (t)-[y:RESPONSE_TO]->(q)
            RETURN t """,
            "t", params={
                "date": date,
                "request_uid": request_uid})

    def get_all_tldrs(self) -> list[Tldr]:
        tldrs = Tldr.nodes
        return tldrs

    def delete_tldr(self, tldr:Tldr):
        self.delete_tldr_entries_by_tldr(tldr)
        tldr.delete()

    def delete_all_tldrs(self):
        for tldr in Tldr.nodes:
            self.delete_tldr(tldr)

    # -----------------
    # Tldr Entries
    # -----------------

    def add_entry_to_tldr(self, tldr:Tldr, score:float, recommendation:Recommendation, summary:Summary, content:Content) -> TldrEntry:
        tldr_entry=TldrEntry(link=content.url,
                            title=content.title,
                            type=content.type,
                            score=recommendation.score,
                            summary=summary.text,
                             ).save()
        tldr_entry.includes.connect(summary)
        tldr_entry.based_on.connect(recommendation)
        tldr.contains.connect(tldr_entry, {'score': score })
        return tldr_entry

    def get_tldr_entry_by_uid(self, uid:str) -> Tldr:
        tldr_entry = TldrEntry.nodes.get_or_none(uid=uid)
        return tldr_entry

    def get_entries_by_tldr(self, tldr:Tldr) -> list[TldrEntry]:
        return self.cypher_query("""
            MATCH (t:Tldr) WHERE t.uid=$tldr_id
            MATCH (e:TldrEntry)
            MATCH (t)-[x:CONTAINS]->(e)
            RETURN e ORDER BY e.score DESC""",
            "e", params={"tldr_id": tldr.uid})

    def delete_tldr_entry(self, tldr_entry:TldrEntry):
        tldr_entry.delete()

    def delete_tldr_entries_by_tldr(self, tldr:Tldr):
        for tldr_entry in tldr.contains.all():
            self.delete_tldr_entry(tldr_entry)

	#-----------------
	# Feedback
	#-----------------

    def add_feedback_click(self, entry:TldrEntry, date:datetime.date) -> Feedback:
        feedback = self.get_feedback_by_tldr_entry(entry)
        if feedback is None:
            feedback= Feedback(click=date).save()
            feedback.about_entry.connect(entry)
            request= self.get_requests_by_tldr_entry(entry)
            feedback.from_request.connect(request)
        else:
            feedback.click_date=date
            feedback.save()
            log.warning("Updated click date of existing Feedback.")

        return feedback

    def add_feedback_rating(self, entry:TldrEntry, score:float) -> Feedback:
        feedback = self.get_feedback_by_tldr_entry(entry)
        if feedback is None:
            feedback= Feedback(score=score).save()
            feedback.about_entry.connect(entry)
            request= self.get_requests_by_tldr_entry(entry)
            feedback.from_request.connect(request)
        else:
            feedback.score=score
            feedback.save()
            log.warning("Updated scoring of existing Feedback.")

        return feedback

    def get_feedback_by_tldr_entry(self, entry:TldrEntry) -> Feedback:
         return self.cypher_query_one("""
            MATCH (f:Feedback)
            MATCH (e:TldrEntry) WHERE e.uid=$uid
            MATCH (f)-[x:ABOUT]->(e)
            RETURN f """, "f", params={"uid":entry.uid})     
    
    def get_feedback_by_request(self, request:Request) -> list[Feedback]:
         return self.cypher_query("""
            MATCH (f:Feedback)
            MATCH (r:Request) WHERE e.uid=$uid
            MATCH (f)-[x:FROM]->(r)
            RETURN f """, "f", params={"uid":request.uid})

	#-----------------
	# EvalKey Nodes - rubric for evaluation metrics
	#-----------------


    def add_evalkey(self, content:Content, request:Request, score:float, text:str) -> EvalKey:
        key = EvalKey(text=text, score=score).save()
        key.key_for_content.connect(content)
        key.key_for_request.connect(request)
        return key

    def get_evalkey_by_uid(self, uid: str) -> EvalKey:
        key = EvalKey.nodes.get_or_none(uid=uid)
        return key

    def get_evalkeys_by_request(self, request: Request) -> list[EvalKey]:
        return self.cypher_query("""
            MATCH (q:Request) WHERE q.uid=$request_id
            MATCH (e:EvalKey)-[eq:KEY_FOR_REQUEST]->(q)
            RETURN e """,
            "e", params={"request_id": request.uid})

    def get_evalkeys_by_content(self, content: Content) -> list[EvalKey]:
        return self.cypher_query("""
            MATCH (c:Content) WHERE c.uid=$content_id
            MATCH (c)<-[ec:KEY_FOR_CONTENT]-(e:EvalKey)
            RETURN e """,
            "e", params={"content_id": content.uid})

    def get_evalkey_by_content_and_request(self, content: Content, request: Request) -> EvalKey:
        return self.cypher_query_one("""
            MATCH (c:Content) WHERE c.uid=$content_id
            MATCH (q:Request) WHERE q.uid=$request_id
            MATCH (c)<-[ec:KEY_FOR_CONTENT]-(e:EvalKey)-[eq:KEY_FOR_REQUEST]->(q)
            RETURN e """,
            "e", params={
                "content_id": content.uid,
                "request_id": request.uid})

    def get_all_evalkeys(self) -> list[EvalKey]:
        evalkeys = EvalKey.nodes
        return evalkeys

    def delete_evalkey(self, evalKey:EvalKey):
        return self.delete_evalkey_by_uid(evalKey.uid)

    def delete_evalkey_by_uid(self, uid:str):
        return self.cypher_query("""
            MATCH (k:EvalKey) WHERE k.uid=$uid
            DETACH DELETE k""",
            params={"uid": uid})

    def delete_all_evalkeys(self):
        for k in self.get_all_evalkeys():
            self.delete_evalkey(k)
